# -*- coding: utf-8 -*-

#
# IceCream - Never use print() to debug again
#
# Ansgar Grunseid
# grunseid.com
# grunseid@gmail.com
#
# License: MIT
#

__title__ = 'icecream'
__license__ = 'MIT'
__version__ = '2.1.3'
__author__ = 'Ansgar Grunseid'
__contact__ = 'grunseid@gmail.com'
__url__ = 'https://github.com/gruns/icecream'
__description__ = (
    'Never use print() to debug again; inspect variables, expressions, and '
    'program execution with a single, simple function call.')
